/*
 * Copyright (c) 2007 Henri Sivonen
 * Copyright (c) 2008-2011 Mozilla Foundation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * THIS IS A GENERATED FILE. PLEASE DO NOT EDIT.
 * Please edit HtmlAttributes.java instead and regenerate.
 */

mod Tokenizer;
mod TreeBuilder;
mod MetaScanner;
mod AttributeName;
mod ElementName;
mod StackNode;
mod UTF16Buffer;
mod StateSnapshot;
/**
 * Be careful with this class. QName is the name in from HTML tokenization.
 * Otherwise, please refer to the interface doc.
 * 
 * @version $Id: AttributesImpl.java 206 2008-03-20 14:09:29Z hsivonen $
 * @author hsivonen
 */

pub const EMPTY_ATTRIBUTES: @HtmlAttributes = new HtmlAttributes(AttributeName::HTML);


struct HtmlAttributes {

    mut mode: i32,

    mut length: i32,

    mut names: @[AttributeName],

    mut values: @[String],
}

impl HtmlAttributes {

    fn destructor() {
        clear(0);
    }

    /**
     * Only use with a static argument
     * 
     * @param name
     * @return
     */
    fn getIndex(name: @AttributeName) -> i32 {
        let mut i: i32 = 0;
        while i < self.length {
            if self.names[i] == name {
                return i;
            }
            i = i + 1;
        }
        return -1;
    }

    /**
     * Only use with static argument.
     * 
     * @see org.xml.sax.Attributes#getValue(java.lang.String)
     */
    fn getValue(name: @AttributeName) -> @String {
        let mut index: i32 = getIndex(name);
        if index == -1 {
            return null;
        } else {
            return getValueNoBoundsCheck(index);
        }
    }

    fn getLength() -> i32 {
        return self.length;
    }

    /**
     * Variant of <code>getLocalName(int index)</code> without bounds check.
     * @param index a valid attribute index
     * @return the local name at index
     */
    fn getLocalNameNoBoundsCheck(index: i32) -> @String {
        assert index < self.length && index >= 0;
        return self.names[index].getLocal(self.mode);
    }

    /**
     * Variant of <code>getURI(int index)</code> without bounds check.
     * @param index a valid attribute index
     * @return the namespace URI at index
     */
    fn getURINoBoundsCheck(index: i32) -> @String {
        assert index < self.length && index >= 0;
        return self.names[index].getUri(self.mode);
    }

    /**
     * Variant of <code>getPrefix(int index)</code> without bounds check.
     * @param index a valid attribute index
     * @return the namespace prefix at index
     */
    fn getPrefixNoBoundsCheck(index: i32) -> @String {
        assert index < self.length && index >= 0;
        return self.names[index].getPrefix(self.mode);
    }

    /**
     * Variant of <code>getValue(int index)</code> without bounds check.
     * @param index a valid attribute index
     * @return the attribute value at index
     */
    fn getValueNoBoundsCheck(index: i32) -> @String {
        assert index < self.length && index >= 0;
        return self.values[index];
    }

    /**
     * Variant of <code>getAttributeName(int index)</code> without bounds check.
     * @param index a valid attribute index
     * @return the attribute name at index
     */
    fn getAttributeNameNoBoundsCheck(index: i32) -> @AttributeName {
        assert index < self.length && index >= 0;
        return self.names[index];
    }

    fn addAttribute(name: @AttributeName, value: @String) {
        if names.len() as i32 == self.length {
            let mut newLen: i32 = self.length << 1;
            let mut newNames: @[AttributeName] = new AttributeName[newLen];
            System.arraycopy(self.names, 0, newNames, 0, names.len() as i32);
            self.names = newNames;
            let mut newValues: @[String] = new String[newLen];
            System.arraycopy(self.values, 0, newValues, 0, values.len() as i32);
            self.values = newValues;
        }
        self.names[self.length] = name;
        self.values[self.length] = value;
        self.length = self.length + 1;
    }

    fn clear(m: i32) {
        let mut i: i32 = 0;
        while i < self.length {
            self.names[i].release();
            self.names[i] = null;
            Portability.releaseString(self.values[i]);
            self.values[i] = null;
            i = i + 1;
        }
        self.length = 0;
        self.mode = m;
    }

    /**
     * This is used in C++ to release special <code>isindex</code>
     * attribute values whose ownership is not transferred.
     */
    fn releaseValue(i: i32) {
        Portability.releaseString(self.values[i]);
    }

    /**
     * This is only used for <code>AttributeName</code> ownership transfer
     * in the isindex case to avoid freeing custom names twice in C++.
     */
    fn clearWithoutReleasingContents() {
        let mut i: i32 = 0;
        while i < self.length {
            self.names[i] = null;
            self.values[i] = null;
            i = i + 1;
        }
        self.length = 0;
    }

    fn contains(name: @AttributeName) -> bool {
        let mut i: i32 = 0;
        while i < self.length {
            if name.equalsAnother(self.names[i]) {
                return true;
            }
            i = i + 1;
        }
        return false;
    }

    fn adjustForMath() {
        self.mode = AttributeName::MATHML;
    }

    fn adjustForSvg() {
        self.mode = AttributeName::SVG;
    }

    fn cloneAttributes(interner: @Interner) -> @HtmlAttributes {
        assert (self.length == 0) || self.mode == 0 || self.mode == 3;
        let mut clone: @HtmlAttributes = new HtmlAttributes(0);
        let mut i: i32 = 0;
        while i < self.length {
            clone.addAttribute(self.names[i].cloneAttributeName(interner), Portability.newStringFromString(self.values[i]));
            i = i + 1;
        }
        return clone;
    }

    fn equalsAnother(other: @HtmlAttributes) -> bool {
        assert self.mode == 0 || self.mode == 3;
        let mut otherLength: i32 = other.getLength();
        if self.length != otherLength {
            return false;
        }
        let mut i: i32 = 0;
        while i < self.length {
            let mut found: bool = false;
            let mut ownLocal: @String = self.names[i].getLocal(AttributeName::HTML);
            let mut j: i32 = 0;
            while j < otherLength {
                if ownLocal == other.names[j].getLocal(AttributeName::HTML) {
                    found = true;
                    if !Portability.stringEqualsString(self.values[i], other.values[j]) {
                        return false;
                    }
                }
                j = j + 1;
            }
            if !found {
                return false;
            }
            i = i + 1;
        }
        return true;
    }
}
