/*
 * Copyright (c) 2007 Henri Sivonen
 * Copyright (c) 2008-2010 Mozilla Foundation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * THIS IS A GENERATED FILE. PLEASE DO NOT EDIT.
 * Please edit MetaScanner.java instead and regenerate.
 */

mod Tokenizer;
mod TreeBuilder;
mod AttributeName;
mod ElementName;
mod HtmlAttributes;
mod StackNode;
mod UTF16Buffer;
mod StateSnapshot;

/**
     * Constant for "charset".
     */
const CHARSET: @[u16] = [ 0x68, 0x61, 0x72, 0x73, 0x65, 0x74 ];

/**
     * Constant for "content".
     */
const CONTENT: @[u16] = [ 0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74 ];

/**
     * Constant for "http-equiv".
     */
const HTTP_EQUIV: @[u16] = [ 0x74, 0x74, 0x70, 0x2d, 0x65, 0x71, 0x75, 0x69, 0x76 ];

/**
     * Constant for "content-type".
     */
const CONTENT_TYPE: @[u16] = [ 0x63, 0x6f, 0x6e, 0x74, 0x65, 0x6e, 0x74, 0x2d, 0x74, 0x79, 0x70, 0x65 ];

const NO: i32 = 0;

const M: i32 = 1;

const E: i32 = 2;

const T: i32 = 3;

const A: i32 = 4;

const DATA: i32 = 0;

const TAG_OPEN: i32 = 1;

const SCAN_UNTIL_GT: i32 = 2;

const TAG_NAME: i32 = 3;

const BEFORE_ATTRIBUTE_NAME: i32 = 4;

const ATTRIBUTE_NAME: i32 = 5;

const AFTER_ATTRIBUTE_NAME: i32 = 6;

const BEFORE_ATTRIBUTE_VALUE: i32 = 7;

const ATTRIBUTE_VALUE_DOUBLE_QUOTED: i32 = 8;

const ATTRIBUTE_VALUE_SINGLE_QUOTED: i32 = 9;

const ATTRIBUTE_VALUE_UNQUOTED: i32 = 10;

const AFTER_ATTRIBUTE_VALUE_QUOTED: i32 = 11;

const MARKUP_DECLARATION_OPEN: i32 = 13;

const MARKUP_DECLARATION_HYPHEN: i32 = 14;

const COMMENT_START: i32 = 15;

const COMMENT_START_DASH: i32 = 16;

const COMMENT: i32 = 17;

const COMMENT_END_DASH: i32 = 18;

const COMMENT_END: i32 = 19;

const SELF_CLOSING_START_TAG: i32 = 20;

const HTTP_EQUIV_NOT_SEEN: i32 = 0;

const HTTP_EQUIV_CONTENT_TYPE: i32 = 1;

const HTTP_EQUIV_OTHER: i32 = 2;


struct MetaScanner {

    /**
     * The data source.
     */
    mut readable: @ByteReadable,

    /**
     * The state of the state machine that recognizes the tag name "meta".
     */
    mut metaState: i32,

    /**
     * The current position in recognizing the attribute name "content".
     */
    mut contentIndex: i32,

    /**
     * The current position in recognizing the attribute name "charset".
     */
    mut charsetIndex: i32,

    /**
     * The current position in recognizing the attribute name "http-equive".
     */
    mut httpEquivIndex: i32,

    /**
     * The current position in recognizing the attribute value "content-type".
     */
    mut contentTypeIndex: i32,

    /**
     * The tokenizer state.
     */
    mut stateSave: i32,

    /**
     * The currently filled length of strBuf.
     */
    mut strBufLen: i32,

    /**
     * Accumulation buffer for attribute values.
     */
    mut strBuf: @[u16],

    mut content: @String,

    mut charset: @String,

    mut httpEquivState: i32,
}

impl MetaScanner {

    fn destructor() {
        Portability.releaseString(self.content);
        Portability.releaseString(self.charset);
    }

    /**
     * The runs the meta scanning algorithm.
     */
    fn stateLoop(state: i32) {
        let mut c: i32 = -1;
        let mut reconsume: bool = false;
        loop stateloop: {
            match state {
                DATA => {
                    loop dataloop: {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x3c => {
                                state = MetaScanner::TAG_OPEN;
                                break dataloop;
                            }
                            _ => {
                                loop;
                            }
                        }
                    }
                }
                TAG_OPEN => {
                    loop tagopenloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x6d | 0x4d => {
                                self.metaState = M;
                                state = MetaScanner::TAG_NAME;
                                break tagopenloop;
                            }
                            0x21 => {
                                state = MetaScanner::MARKUP_DECLARATION_OPEN;
                                loop stateloop;
                            }
                            0x3f | 0x2f => {
                                state = MetaScanner::SCAN_UNTIL_GT;
                                loop stateloop;
                            }
                            0x3e => {
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                if (c >= 0x41 && c <= 0x5a) || (c >= 0x61 && c <= 0x7a) {
                                    self.metaState = NO;
                                    state = MetaScanner::TAG_NAME;
                                    break tagopenloop;
                                }
                                state = MetaScanner::DATA;
                                reconsume = true;
                                loop stateloop;
                            }
                        }
                    }
                }
                TAG_NAME => {
                    loop tagnameloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                state = MetaScanner::BEFORE_ATTRIBUTE_NAME;
                                break tagnameloop;
                            }
                            0x2f => {
                                state = MetaScanner::SELF_CLOSING_START_TAG;
                                loop stateloop;
                            }
                            0x3e => {
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            0x65 | 0x45 => {
                                if self.metaState == M {
                                    self.metaState = E;
                                } else {
                                    self.metaState = NO;
                                }
                                loop;
                            }
                            0x74 | 0x54 => {
                                if self.metaState == E {
                                    self.metaState = T;
                                } else {
                                    self.metaState = NO;
                                }
                                loop;
                            }
                            0x61 | 0x41 => {
                                if self.metaState == T {
                                    self.metaState = A;
                                } else {
                                    self.metaState = NO;
                                }
                                loop;
                            }
                            _ => {
                                self.metaState = NO;
                                loop;
                            }
                        }
                    }
                }
                BEFORE_ATTRIBUTE_NAME => {
                    loop beforeattributenameloop: {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                loop;
                            }
                            0x2f => {
                                state = MetaScanner::SELF_CLOSING_START_TAG;
                                loop stateloop;
                            }
                            0x3e => {
                                if handleTag() {
                                    break stateloop;
                                }
                                state = DATA;
                                loop stateloop;
                            }
                            0x63 | 0x43 => {
                                self.contentIndex = 0;
                                self.charsetIndex = 0;
                                self.httpEquivIndex = Integer.MAX_VALUE;
                                self.contentTypeIndex = Integer.MAX_VALUE;
                                state = MetaScanner::ATTRIBUTE_NAME;
                                break beforeattributenameloop;
                            }
                            0x68 | 0x48 => {
                                self.contentIndex = Integer.MAX_VALUE;
                                self.charsetIndex = Integer.MAX_VALUE;
                                self.httpEquivIndex = 0;
                                self.contentTypeIndex = Integer.MAX_VALUE;
                                state = MetaScanner::ATTRIBUTE_NAME;
                                break beforeattributenameloop;
                            }
                            _ => {
                                self.contentIndex = Integer.MAX_VALUE;
                                self.charsetIndex = Integer.MAX_VALUE;
                                self.httpEquivIndex = Integer.MAX_VALUE;
                                self.contentTypeIndex = Integer.MAX_VALUE;
                                state = MetaScanner::ATTRIBUTE_NAME;
                                break beforeattributenameloop;
                            }
                        }
                    }
                }
                ATTRIBUTE_NAME => {
                    loop attributenameloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                state = MetaScanner::AFTER_ATTRIBUTE_NAME;
                                loop stateloop;
                            }
                            0x2f => {
                                state = MetaScanner::SELF_CLOSING_START_TAG;
                                loop stateloop;
                            }
                            0x3d => {
                                self.strBufLen = 0;
                                self.contentTypeIndex = 0;
                                state = MetaScanner::BEFORE_ATTRIBUTE_VALUE;
                                break attributenameloop;
                            }
                            0x3e => {
                                if handleTag() {
                                    break stateloop;
                                }
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                if self.metaState == A {
                                    if c >= 0x41 && c <= 0x5a {
                                        c += 0x20;
                                    }
                                    if self.contentIndex < CONTENT.len() as i32 && c == CONTENT[self.contentIndex] {
                                        self.contentIndex = self.contentIndex + 1;
                                    } else {
                                        self.contentIndex = Integer.MAX_VALUE;
                                    }
                                    if self.charsetIndex < CHARSET.len() as i32 && c == CHARSET[self.charsetIndex] {
                                        self.charsetIndex = self.charsetIndex + 1;
                                    } else {
                                        self.charsetIndex = Integer.MAX_VALUE;
                                    }
                                    if self.httpEquivIndex < HTTP_EQUIV.len() as i32 && c == HTTP_EQUIV[self.httpEquivIndex] {
                                        self.httpEquivIndex = self.httpEquivIndex + 1;
                                    } else {
                                        self.httpEquivIndex = Integer.MAX_VALUE;
                                    }
                                }
                                loop;
                            }
                        }
                    }
                }
                BEFORE_ATTRIBUTE_VALUE => {
                    loop beforeattributevalueloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                loop;
                            }
                            0x22 => {
                                state = MetaScanner::ATTRIBUTE_VALUE_DOUBLE_QUOTED;
                                break beforeattributevalueloop;
                            }
                            0x27 => {
                                state = MetaScanner::ATTRIBUTE_VALUE_SINGLE_QUOTED;
                                loop stateloop;
                            }
                            0x3e => {
                                if handleTag() {
                                    break stateloop;
                                }
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                handleCharInAttributeValue(c);
                                state = MetaScanner::ATTRIBUTE_VALUE_UNQUOTED;
                                loop stateloop;
                            }
                        }
                    }
                }
                ATTRIBUTE_VALUE_DOUBLE_QUOTED => {
                    loop attributevaluedoublequotedloop: {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x22 => {
                                handleAttributeValue();
                                state = MetaScanner::AFTER_ATTRIBUTE_VALUE_QUOTED;
                                break attributevaluedoublequotedloop;
                            }
                            _ => {
                                handleCharInAttributeValue(c);
                                loop;
                            }
                        }
                    }
                }
                AFTER_ATTRIBUTE_VALUE_QUOTED => {
                    loop afterattributevaluequotedloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                state = MetaScanner::BEFORE_ATTRIBUTE_NAME;
                                loop stateloop;
                            }
                            0x2f => {
                                state = MetaScanner::SELF_CLOSING_START_TAG;
                                break afterattributevaluequotedloop;
                            }
                            0x3e => {
                                if handleTag() {
                                    break stateloop;
                                }
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                state = MetaScanner::BEFORE_ATTRIBUTE_NAME;
                                reconsume = true;
                                loop stateloop;
                            }
                        }
                    }
                }
                SELF_CLOSING_START_TAG => {
                    c = read();
                    match c {
                        -1 => {
                            break stateloop;
                        }
                        0x3e => {
                            if handleTag() {
                                break stateloop;
                            }
                            state = MetaScanner::DATA;
                            loop stateloop;
                        }
                        _ => {
                            state = MetaScanner::BEFORE_ATTRIBUTE_NAME;
                            reconsume = true;
                            loop stateloop;
                        }
                    }
                }
                ATTRIBUTE_VALUE_UNQUOTED => {
                    loop {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                handleAttributeValue();
                                state = MetaScanner::BEFORE_ATTRIBUTE_NAME;
                                loop stateloop;
                            }
                            0x3e => {
                                handleAttributeValue();
                                if handleTag() {
                                    break stateloop;
                                }
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                handleCharInAttributeValue(c);
                                loop;
                            }
                        }
                    }
                }
                AFTER_ATTRIBUTE_NAME => {
                    loop {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x20 | 0x09 | 0x0A | 0x0c => {
                                loop;
                            }
                            0x2f => {
                                handleAttributeValue();
                                state = MetaScanner::SELF_CLOSING_START_TAG;
                                loop stateloop;
                            }
                            0x3d => {
                                self.strBufLen = 0;
                                self.contentTypeIndex = 0;
                                state = MetaScanner::BEFORE_ATTRIBUTE_VALUE;
                                loop stateloop;
                            }
                            0x3e => {
                                handleAttributeValue();
                                if handleTag() {
                                    break stateloop;
                                }
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            0x63 | 0x43 => {
                                self.contentIndex = 0;
                                self.charsetIndex = 0;
                                state = MetaScanner::ATTRIBUTE_NAME;
                                loop stateloop;
                            }
                            _ => {
                                self.contentIndex = Integer.MAX_VALUE;
                                self.charsetIndex = Integer.MAX_VALUE;
                                state = MetaScanner::ATTRIBUTE_NAME;
                                loop stateloop;
                            }
                        }
                    }
                }
                MARKUP_DECLARATION_OPEN => {
                    loop markupdeclarationopenloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x2d => {
                                state = MetaScanner::MARKUP_DECLARATION_HYPHEN;
                                break markupdeclarationopenloop;
                            }
                            _ => {
                                state = MetaScanner::SCAN_UNTIL_GT;
                                reconsume = true;
                                loop stateloop;
                            }
                        }
                    }
                }
                MARKUP_DECLARATION_HYPHEN => {
                    loop markupdeclarationhyphenloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x2d => {
                                state = MetaScanner::COMMENT_START;
                                break markupdeclarationhyphenloop;
                            }
                            _ => {
                                state = MetaScanner::SCAN_UNTIL_GT;
                                reconsume = true;
                                loop stateloop;
                            }
                        }
                    }
                }
                COMMENT_START => {
                    loop commentstartloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x2d => {
                                state = MetaScanner::COMMENT_START_DASH;
                                loop stateloop;
                            }
                            0x3e => {
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                state = MetaScanner::COMMENT;
                                break commentstartloop;
                            }
                        }
                    }
                }
                COMMENT => {
                    loop commentloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x2d => {
                                state = MetaScanner::COMMENT_END_DASH;
                                break commentloop;
                            }
                            _ => {
                                loop;
                            }
                        }
                    }
                }
                COMMENT_END_DASH => {
                    loop commentenddashloop: {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x2d => {
                                state = MetaScanner::COMMENT_END;
                                break commentenddashloop;
                            }
                            _ => {
                                state = MetaScanner::COMMENT;
                                loop stateloop;
                            }
                        }
                    }
                }
                COMMENT_END => {
                    loop {
                        c = read();
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x3e => {
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            0x2d => {
                                loop;
                            }
                            _ => {
                                state = MetaScanner::COMMENT;
                                loop stateloop;
                            }
                        }
                    }
                }
                COMMENT_START_DASH => {
                    c = read();
                    match c {
                        -1 => {
                            break stateloop;
                        }
                        0x2d => {
                            state = MetaScanner::COMMENT_END;
                            loop stateloop;
                        }
                        0x3e => {
                            state = MetaScanner::DATA;
                            loop stateloop;
                        }
                        _ => {
                            state = MetaScanner::COMMENT;
                            loop stateloop;
                        }
                    }
                }
                ATTRIBUTE_VALUE_SINGLE_QUOTED => {
                    loop {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x27 => {
                                handleAttributeValue();
                                state = MetaScanner::AFTER_ATTRIBUTE_VALUE_QUOTED;
                                loop stateloop;
                            }
                            _ => {
                                handleCharInAttributeValue(c);
                                loop;
                            }
                        }
                    }
                }
                SCAN_UNTIL_GT => {
                    loop {
                        if reconsume {
                            reconsume = false;
                        } else {
                            c = read();
                        }
                        match c {
                            -1 => {
                                break stateloop;
                            }
                            0x3e => {
                                state = MetaScanner::DATA;
                                loop stateloop;
                            }
                            _ => {
                                loop;
                            }
                        }
                    }
                }
            }
        }
        self.stateSave = state;
    }

    fn handleCharInAttributeValue(c: i32) {
        if self.metaState == A {
            if self.contentIndex == CONTENT.len() as i32 || self.charsetIndex == CHARSET.len() as i32 {
                addToBuffer(c);
            } else if self.httpEquivIndex == HTTP_EQUIV.len() as i32 {
                if self.contentTypeIndex < CONTENT_TYPE.len() as i32 && toAsciiLowerCase(c) == CONTENT_TYPE[self.contentTypeIndex] {
                    self.contentTypeIndex = self.contentTypeIndex + 1;
                } else {
                    self.contentTypeIndex = Integer.MAX_VALUE;
                }
            }
        }
    }

    fn toAsciiLowerCase(c: i32) -> i32 {
        if c >= 0x41 && c <= 0x5a {
            return c + 0x20;
        }
        return c;
    }

    /**
     * Adds a character to the accumulation buffer.
     * @param c the character to add
     */
    fn addToBuffer(c: i32) {
        if self.strBufLen == strBuf.len() as i32 {
            let mut newBuf: @[u16] = new u16[strBuf.len() as i32 + (strBuf.len() as i32 << 1)];
            System.arraycopy(self.strBuf, 0, newBuf, 0, strBuf.len() as i32);
            self.strBuf = newBuf;
        }
        self.strBuf[self.strBufLen] = (u16) c;
        self.strBufLen = self.strBufLen + 1;
    }

    /**
     * Attempts to extract a charset name from the accumulation buffer.
     * @return <code>true</code> if successful
     * @throws SAXException
     */
    fn handleAttributeValue() {
        if self.metaState != A {
            return;
        }
        if self.contentIndex == CONTENT.len() as i32 && self.content == null {
            self.content = Portability.newStringFromBuffer(self.strBuf, 0, self.strBufLen);
            return;
        }
        if self.charsetIndex == CHARSET.len() as i32 && self.charset == null {
            self.charset = Portability.newStringFromBuffer(self.strBuf, 0, self.strBufLen);
            return;
        }
        if self.httpEquivIndex == HTTP_EQUIV.len() as i32 && self.httpEquivState == HTTP_EQUIV_NOT_SEEN {
            self.httpEquivState = (self.contentTypeIndex == CONTENT_TYPE.len() as i32) ? HTTP_EQUIV_CONTENT_TYPE : HTTP_EQUIV_OTHER;
            return;
        }
    }

    fn handleTag() -> bool {
        let mut stop: bool = handleTagInner();
        Portability.releaseString(self.content);
        self.content = null;
        Portability.releaseString(self.charset);
        self.charset = null;
        self.httpEquivState = HTTP_EQUIV_NOT_SEEN;
        return stop;
    }

    fn handleTagInner() -> bool {
        if self.charset != null && tryCharset(self.charset) {
            return true;
        }
        if self.content != null && self.httpEquivState == HTTP_EQUIV_CONTENT_TYPE {
            let mut extract: @String = TreeBuilder.extractCharsetFromContent(self.content);
            if extract == null {
                return false;
            }
            let mut success: bool = tryCharset(extract);
            Portability.releaseString(extract);
            return success;
        }
        return false;
    }

    /**
     * Tries to switch to an encoding.
     * 
     * @param encoding
     * @return <code>true</code> if successful
     * @throws SAXException
     */
    fn tryCharset(encoding: @String) -> bool;
}
