/*
 * Copyright (c) 2008-2010 Mozilla Foundation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * THIS IS A GENERATED FILE. PLEASE DO NOT EDIT.
 * Please edit UTF16Buffer.java instead and regenerate.
 */

mod Tokenizer;
mod TreeBuilder;
mod MetaScanner;
mod AttributeName;
mod ElementName;
mod HtmlAttributes;
mod StackNode;
mod StateSnapshot;
/**
 * An UTF-16 buffer that knows the start and end indeces of its unconsumed
 * content.
 * 
 * @version $Id$
 * @author hsivonen
 */


struct UTF16Buffer {

    /**
     * The backing store of the buffer. May be larger than the logical content
     * of this <code>UTF16Buffer</code>.
     */
    buffer: @[u16],

    /**
     * The index of the first unconsumed character in the backing buffer.
     */
    mut start: i32,

    /**
     * The index of the slot immediately after the last character in the backing
     * buffer that is part of the logical content of this
     * <code>UTF16Buffer</code>.
     */
    mut end: i32,
}

impl UTF16Buffer {

    /**
     * Returns the start index.
     * 
     * @return the start index
     */
    fn getStart() -> i32 {
        return self.start;
    }

    /**
     * Sets the start index.
     * 
     * @param start
     *            the start index
     */
    fn setStart(start: i32) {
        this.start = self.start;
    }

    /**
     * Returns the backing buffer.
     * 
     * @return the backing buffer
     */
    fn getBuffer() -> @[u16] {
        return self.buffer;
    }

    /**
     * Returns the end index.
     * 
     * @return the end index
     */
    fn getEnd() -> i32 {
        return self.end;
    }

    /**
     * Checks if the buffer has data left.
     * 
     * @return <code>true</code> if there's data left
     */
    fn hasMore() -> bool {
        return self.start < self.end;
    }

    /**
     * Adjusts the start index to skip over the first character if it is a line
     * feed and the previous character was a carriage return.
     * 
     * @param lastWasCR
     *            whether the previous character was a carriage return
     */
    fn adjust(lastWasCR: bool) {
        if lastWasCR && self.buffer[self.start] == 0x0A {
            self.start = self.start + 1;
        }
    }

    /**
     * Sets the end index.
     * 
     * @param end
     *            the end index
     */
    fn setEnd(end: i32) {
        this.end = self.end;
    }
}
